/*  ftpd.c - Minimal FTP server */

#include <stdio.h>
#include <string.h>
#include "types.h"
#include "interrupts.h"
#include "fiber.h"
#include "cpmio.h"
#include "net.h"
#include "log.h"
#include "wiz.h"
#include "w5100.h"

#define FTPC_BLEN   80              // Command buffer length
#define FTPD_BLEN   128             // Data buffer length
#define PORT_FTPD   20              // Default data port
#define PORT_FTPC   21              // Command port

#define mdIdle  0x00                // Data connection idle
#define mdSend  0x10                // Send file
#define mdRecv  0x20                // Receive file
#define mdList  0x30                // List directory
#define mdCmnd  0xF0                // Mask to select action
#define mdLEnd  0x01                // End of listing
#define mdLDir  0x04                // List attributes (LIST)
#define mdMLst  0x08                // Machine listing (MLSD)

#define DIRMAX  512                 // Maximum number of files in directory

typedef struct
    {
    FIBER           fbr;            // Fibre for enquing actions
    SOCKET_ALLOC    sacmd;          // Command Socket allocation
    SOCKET_ALLOC    sadat;          // Data Socket allocation
    char            buf[FTPC_BLEN]; // Command buffer
    const char *    psta;           // Pointer to status;
    int             rptr;           // Read position in command buffer
    int             wptr;           // Write position in command buffer
    byte            daddr[6];       // Destination IP address and port
    word            pasv;           // Port for passive transfers (or zero if not passive)
    byte            md;             // Transfer mode
    byte            dcon;           // Data connection established
    int             iFind;          // Index of file to find
    FCB             fcb;            // FCB for file to transfer
    byte            dat[FTPD_BLEN]; // Data buffer
    int             ndat;           // Number of data bytes received
    } FTP;

FTP ftp1;
FTP ftp2;

typedef struct
    {
    char            name[11];       // File name and extension
    unsigned int    nkb;            // Number of KB
    } DITEM;

DITEM   dlist[DIRMAX];              // Names and sizes of all files
unsigned int nlist = 0;             // Number of files
unsigned int nLoop = 0;

static char sta150[] = "150 Starting transfer\r\n";
static char sta200[] = "200 OK\r\n";
static char sta211[] = "211- Features\r\n MLST type*;size*;\r\n211 End\r\n";
static char sta215[] = "215 CPM 2.2 Memotech MTX\r\n";
static char sta220[] = "220 Ready\r\n";
static char sta221[] = "221 Logoff\r\n";
static char sta226[] = "226 Success / Closing\r\n";
static char sta227[] = "227 Passive mode .???,???,???,???,???,???.\r\n";
static char sta230[] = "230 Logon\r\n";
static char sta250[] = "250 File OK\r\n";
static char sta257[] = "257 \"?:/\" Current drive\r\n";
static char sta350[] = "350 More info\r\n";
static char sta500[] = "500 Unrecognised command\r\n";
static char sta501[] = "501 Invalid parameters\r\n";
static char sta504[] = "504 Option not implemented\r\n";
static char sta550[] = "550 File not found\r\n";
static char sta552[] = "552 No space\r\n";

static const char *psStatus[] = {
    " SS_FREE\r\n$",
    " SS_ALLOCATED\r\n$",
    " SS_CONNECT\r\n$",
    " SS_LISTEN\r\n$",
    " SS_ESTABLISHED\r\n$",
    " SS_SEND\r\n$",
    " SS_RECV\r\n$",
    " SS_ERROR\r\n$",
    " SS_DISCONNECT\r\n$"
    };

#ifdef DEBUG
static void log_head (SOCKET *socket)
    {
    log_uint (nLoop);
    log_printf (" S$");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (": $");
    }
#endif

static void ftp_status (SOCKET *socket)
    {
    log_printf ("Socket $");
    log_uint ((socket->sra - W5100_ADDR_SOCKET) >> 8);
    log_printf (psStatus[socket->state]);
    }

static void ftpc_recv(FTP *pftp);
static void ftpc_listen(FTP *pftp);
static void ftpc_read(FTP *pftp);
static void ftpd_recv(FTP *pftp);
static void ftpd_connect (FTP *pftp);

static char *ftp_copy (char *psDst, const char *psSrc)
    {
    while (*psSrc)
        {
        *psDst = *psSrc;
        ++psSrc;
        ++psDst;
        }
    *psDst = '\0';
    return psDst;
    }

static void ftp_dgt (char **pps, word *pw, byte *pf, word d)
    {
    char ch = '0';
    while ( *pw >= d )
        {
        *pw -= d;
        ++ch;
        }
    if (( ch > '0' ) || ( *pf == 2 ))
        {
        **pps = ch;
        ++(*pps);
        *pf = 2;
        }
    else if ( *pf == 1 )
        {
        **pps = ' ';
        ++(*pps);
        }
    }

static char *ftp_uint (char *ps, word w)
    {
    byte f = 0;
    ftp_dgt (&ps, &w, &f, 10000);
    ftp_dgt (&ps, &w, &f, 1000);
    ftp_dgt (&ps, &w, &f, 100);
    ftp_dgt (&ps, &w, &f, 10);
    *ps = w + '0';
    return ++ps;
    }

static char *ftp_uintp (char *ps, word w)
    {
    byte f = 1;
    ftp_dgt (&ps, &w, &f, 10000);
    ftp_dgt (&ps, &w, &f, 1000);
    ftp_dgt (&ps, &w, &f, 100);
    ftp_dgt (&ps, &w, &f, 10);
    *ps = w + '0';
    return ++ps;
    }

static BOOLEAN ftp_fname (const char *ps, FCB *pfcb)
    {
    const char *ps0 = ps;
    int n = 0;
    char ch;
    while ( *ps == ' ' ) ++ps;
    if ( ps[1] == ':' )
        {
        ch = *ps;
        if (( ch >= 'a' ) && ( ch <= 'z' )) ch -= 0x20;
        if (( ch < 'A' ) || ( ch >= 'P' ))
            {
            log_printf ("Invalid drive letter\r\n$");
            return FALSE;
            }
        pfcb->drive = ch - 'A' + 1;
        ps += 2;
        }
    else
        {
        pfcb->drive = 0;
        }
    if ( ( *ps == '/' ) || ( *ps == '\\' ) ) ++ps;
    while ( n < 8 )
        {
        ch = *ps;
        if (( ch <= ' ' ) || ( ch == '.' )) break;
        if ( strchr ("<>,.;:=?*[]", ch) != NULL )
            {
            log_printf ("Invalid character in file name\r\n$");
            return FALSE;
            }
        if (( ch >= 'a' ) && ( ch <= 'z' )) ch -= 0x20;
        pfcb->filename[n] = ch;
        ++ps;
        ++n;
        }
#ifdef DEBUG
    log_printf ("Name length = $");
    log_uint (n);
    log_printf ("\r\n$");
#endif
    while ( n < 8 )
        {
        pfcb->filename[n] = ' ';
        ++n;
        }
    n = 0;
    if ( *ps == '.' )
        {
        ++ps;
        while ( n < 3 )
            {
            ch = *ps;
            if ( ch <= ' ' ) break;
            if ( strchr ("<>,.;:=?*[]", ch) != NULL )
                {
                log_printf ("Invalid character in file extension\r\n$");
                return FALSE;
                }
            if (( ch >= 'a' ) && ( ch <= 'z' )) ch -= 0x20;
            pfcb->type[n] = ch;
            ++ps;
            ++n;
            }
        }
#ifdef DEBUG
    log_printf ("Type length = $");
    log_uint (n);
    log_printf ("\r\n$");
#endif
    while ( n < 3 )
        {
        pfcb->type[n] = ' ';
        ++n;
        }
    if ( *ps > ' ' )
        {
        log_printf ("File name too long: pos = $");
        log_uint (ps - ps0);
        log_printf (", ch = $");
        log_byte (*ps);
        log_printf ("\r\n");
        return FALSE;
        }
    pfcb->ex = 0;
    pfcb->s1 = 0;
    pfcb->s2 = 0;
    return TRUE;
    }

static BOOLEAN ftp_match (const char *psSrch, const char *psTest)
    {
    int i;
    for (i = 0; i < 11; ++i)
        {
        if (( *psSrch != '?' ) && ( *psSrch != (*psTest) & 0x7F )) return FALSE;
        }
    return TRUE;
    }

static void ftp_dlist (void)
    {
    static DIRENT dir[4];
    static FCB fcb;
    byte user;
    DPB *pdpb;
    int kbshft;
    BOOLEAN a16;
    int idx;
    log_printf ("Generating file list ... $");
    user = bdos_get_user ();
    pdpb = bdos_disk_parameters ();
    kbshft = pdpb->block_shift - 3;
    a16 = pdpb->alloc_max > 255;
    nlist = 0;
    strcpy (&fcb.drive, "????????????");
    bdos_set_dma_address (dir);
    idx = bdos_search_for_first (&fcb);
    while ( idx < 4 )
        {
        DIRENT *pdir = &dir[idx];
        if ( pdir->drive == user )
            {
            DITEM *plist = dlist;
            DITEM *pitem = NULL;
            int i;
#ifdef DEBUG
            log_strn (pdir->filename, 11);
            log_str (" ");
            log_uint (pdir->ex);
#endif
            for (i = 0; i < nlist; ++i )
                {
                if ( memcmp (pdir->filename, plist->name, 11) == 0 )
                    {
                    pitem = plist;
                    break;
                    }
                ++plist;
                }
            if (( pitem == NULL ) && ( nlist < DIRMAX ))
                {
                pitem = &dlist[nlist];
                memcpy (pitem->name, pdir->filename, 11);
                pitem->nkb = 0;
                ++nlist;
#ifdef DEBUG
                log_str (" nlist = ");
                log_uint (nlist);
#endif
                }
            if ( pitem != NULL )
                {
                unsigned int nblk = 0;
                int i;
                if ( a16 )
                    {
                    for (i = 0; i < 8; ++i)
                        {
                        if ( pdir->alw[i] > 0 ) ++nblk;
                        }
                    }
                else
                    {
                    for (i = 0; i < 16; ++i)
                        {
                        if ( pdir->alb[i] > 0 ) ++nblk;
                        }
                    }
                pitem->nkb += nblk << kbshft;
#ifdef DEBUG
                log_str (" nblk = ");
                log_uint (nblk);
                log_str (" nkb = ");
                log_uint (pitem->nkb);
#endif
                }
#ifdef DEBUG
            log_str ("\r\n");
#endif
            }
        idx = bdos_search_for_next (&fcb);
        }
    log_printf ("done\r\n");
#ifdef DEBUG
    log_str (" nlist = ");
    log_uint (nlist);
    log_str ("\r\n");
#endif
    }

static void ftp_null(FTP *pftp)
    {
    pftp = pftp;
    }

static void ftpc_reply (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
#endif
    log_printf ("ftpc_reply: $");
    log_line (pftp->psta);
    net_send (pftp->sacmd.socket, pftp->psta, strlen (pftp->psta), (NF) ftpc_read, pftp);
    }

static void ftpd_closed (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_closed:\r\n$");
#endif
    pftp->md = mdIdle;
    pftp->pasv = 0;
    pftp->dcon = 0;
    /*
    if ( pftp->pasv > 0 )
        {
        net_listen (pftp->sadat.socket, pftp->pasv, (NF) ftpd_connect, pftp);
        }
    */
    ftpc_reply (pftp);
    // ftpc_read (pftp);
    }

static void ftpd_close (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_close:\r\n$");
#endif
    net_disconnect (pftp->sadat.socket, (NF) ftpd_closed, pftp);
    }

static void ftpd_done (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_done: $");
    log_line (pftp->psta);
#endif
    net_send (pftp->sacmd.socket, pftp->psta, strlen (pftp->psta), (NF) ftpd_close, pftp);
    }

static void ftpd_send (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_send:\r\n$");
#endif
    bdos_set_dma_address (pftp->dat);
    if ( bdos_read_file (&pftp->fcb) == 0 )
        {
        net_send (pftp->sadat.socket, pftp->dat, FTPD_BLEN, (NF) ftpd_send, pftp);
        }
    else
        {
        bdos_close_file (&pftp->fcb);
        pftp->md = 0;
        pftp->psta = sta226;
        ftpd_close (pftp);
        // ftpd_done (pftp);
        }
    }

static void ftpd_list (FTP *pftp)
    {
    static FCB fcb;
    char *ps = (char *) pftp->dat;
    char *ps1;
    char *psDir;
    int n = 8;
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_list:\r\n$");
    log_str (" nlist = ");
    log_uint (nlist);
    log_str ("\r\n");
#endif
    if ( pftp->md & mdLDir ) n = 2;
    else if ( pftp->md & mdMLst ) n = 3;
    while (( n > 0 ) && ( pftp->iFind < nlist ))
        {
        DITEM *pitem = &dlist[pftp->iFind];
#ifdef DEBUG
        log_str ("iFind = ");
        log_uint (pftp->iFind);
        log_str (" name = ");
        log_strn (pitem->name, 11);
#endif
        if ( ftp_match (pftp->fcb.filename, pitem->name) )
            {
            int j;
            char ch;
            int iSize = pitem->nkb;
            psDir = ps;
#ifdef DEBUG
            log_str (" size = ");
            log_uint (iSize);
#endif
            if ( pftp->md & mdMLst )
                {
                ps = ftp_copy (ps, "type=file;size=");
                ps = ftp_uint (ps, iSize);
                ps = ftp_copy (ps, "000; ");
                }
            else if ( pftp->md & mdLDir )
                {
                if ( pitem->name[8] & 0x80 ) ps = ftp_copy (ps, "-r--r--r--");
                else ps = ftp_copy (ps, "-rw-rw-rw-");
                ps = ftp_copy (ps, " 1 0    0    ");
                ps = ftp_uint (ps, iSize);
                ps = ftp_copy (ps, "000 Jan  1  0:00 ");
                }
            ps1 = ps;
            for (j = 0; j < 8; ++j)
                {
                ch = pitem->name[j] & 0x7F;
                *(ps++) = ch;
                if ( ch > ' ' ) ps1 = ps;
                }
            *(ps1++) = '.';
            ps = ps1;
            for (j = 8; j < 11; ++j)
                {
                ch = pitem->name[j] & 0x7F;
                *(ps++) = ch;
                if ( ch > ' ' ) ps1 = ps;
                }
            *(ps1++) = '\r';
            *(ps1++) = '\n';
#ifdef DEBUG
            log_line (psDir);
#endif
            ps = ps1;
            --n;
            }
        ++pftp->iFind;
#ifdef DEBUG
        log_str ("\r\n");
#endif
        }
    if ( pftp->iFind >= nlist )
        {
        // log_printf ("Dir Close\r\n$");
        pftp->md = 0;
        pftp->psta = sta226;
        if ( ps > pftp->dat ) net_send (pftp->sadat.socket, pftp->dat, ps - pftp->dat, (NF) ftpd_close, pftp);
        else ftpd_close (pftp);
        }
    else
        {
        net_send (pftp->sadat.socket, pftp->dat, ps - pftp->dat, (NF) ftpd_list, pftp);
        }
    }

static void ftpd_read (FTP *pftp)
    {
    BOOLEAN bClose = TRUE;
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_read: $");
    ftp_status (pftp->sadat.socket);
#endif
    // log_printf (" actual = $");
    // log_uint (pftp->sadat.socket->actual);
    // log_printf (" ndat = $");
    pftp->psta = sta226;
    if ( pftp->sadat.socket->state != SS_ERROR )
        {
        pftp->ndat += pftp->sadat.socket->actual;
        // log_uint (pftp->ndat);
        bClose = FALSE;
        }
    else
        {
        // log_uint (pftp->ndat);
        // log_printf (" close$");
        if ( pftp->ndat > 0 )
            {
            int i;
            for (i = pftp->ndat; i < FTPD_BLEN; ++i) pftp->dat[i] = 0x1A;
            pftp->ndat = FTPD_BLEN;
            }
        }
    // log_printf ("\r\n$");
    if ( pftp->ndat >= FTPD_BLEN )
        {
#ifdef DEBUG
        log_printf ("Save record\r\n$");
#endif
        bdos_set_dma_address (pftp->dat);
        if ( bdos_write_file (&pftp->fcb) != 0 )
            {
            log_printf ("Save failed\r\n$");
            bdos_close_file (&pftp->fcb);
            pftp->fcb.ex = 0;
            pftp->fcb.s1 = 0;
            pftp->fcb.s2 = 0;
            pftp->fcb.cr = 0;
            bdos_delete_files (&pftp->fcb);
            bClose = TRUE;
            pftp->psta = sta552;
            }
        pftp->ndat = 0;
        }
    if ( bClose )
        {
#ifdef DEBUG
        log_printf ("Close file\r\n$");
#endif
        bdos_close_file (&pftp->fcb);
        pftp->md = 0;
        ftpd_close (pftp);
        }
    else ftpd_recv (pftp);
    }

static void ftpd_recv (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_recv:\r\n$");
#endif
    net_recv (pftp->sadat.socket, &pftp->dat[pftp->ndat], FTPD_BLEN - pftp->ndat, (NF) ftpd_read, pftp);
    }

static void ftpd_connect (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
#endif
    log_printf ("ftpd_connect: $");
    pftp->dcon = 1;
    switch (pftp->md & mdCmnd)
        {
        case mdSend:
            log_printf ("Send\r\n$");
            ftpd_send (pftp);
            break;
        case mdRecv:
            log_printf ("Receive\r\n$");
            ftpd_recv (pftp);
            break;
        case mdList:
            log_printf ("List\r\n$");
            ftpd_list (pftp);
            break;
        default:
            log_printf ("Unknown\r\n$");
            break;
        }
    }

static void ftpd_open (FTP *pftp)
    {
    word port = ( (word) pftp->daddr[4] << 8 ) | pftp->daddr[5];
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_open:\r\n$");
#endif
    net_connect (pftp->sadat.socket, pftp->daddr, port, (NF) ftpd_connect, pftp);
    }

static void ftpd_start (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sadat.socket);
    log_printf ("ftpd_start:\r\n$");
#endif
    pftp->psta = sta150;
    if ( pftp->pasv == 0 )
        {
#ifdef DEBUG
        log_printf ("Queue ftpd_open\r\n$");
#endif
        pftp->fbr.nf = (NF) ftpd_open;
        pftp->fbr.nfp = pftp;
        fiber_enqueue (&(pftp->fbr));
        }
    else if ( pftp->dcon )
        {
#ifdef DEBUG
        log_printf ("Queue ftpd_connect\r\n$");
#endif
        pftp->fbr.nf = (NF) ftpd_connect;
        pftp->fbr.nfp = pftp;
        fiber_enqueue (&(pftp->fbr));
        }
#ifdef DEBUG
    else
        {
        log_printf ("ftpd_start: wait$");
        log_printf (psStatus[pftp->sadat.socket->state]);
        log_printf (": S_SR = $");
        log_byte (wiz_read(pftp->sadat.socket->sra+W5100_OFFSET_SOCKET_SR));
        log_printf (", S_IR = $");
        log_byte (wiz_read(pftp->sadat.socket->sra+W5100_OFFSET_SOCKET_IR));
        log_printf ("\r\n$");
        }
#endif
    }

static void ftpc_close(FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
    log_printf ("ftpc_close:\r\n$");
#endif
    net_disconnect (pftp->sacmd.socket, (NF) ftpc_listen, pftp);
    }

static void ftpc_quit (FTP *pftp)
    {
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
#endif
    log_printf ("ftpc_quit: $");
    log_line (pftp->psta);
    net_send (pftp->sacmd.socket, pftp->psta, strlen (pftp->psta), (NF) ftpc_close, pftp);
    }

int strncasecmp (const char *ps1, const char *ps2, int n)
    {
    while (n > 0)
        {
        char ch1 = *ps1;
        char ch2 = *ps2;
        if ((ch1 >= 'a') && (ch1 <= 'z')) ch1 -= 0x20;
        if ((ch2 >= 'a') && (ch2 <= 'z')) ch2 -= 0x20;
        ch2 -= ch1;
        if ( ch2 != 0 ) return (int) ch2;
        ++ps1;
        ++ps2;
        --n;
        }
    return 0;
    }

static void ftpc_cmnd(FTP *pftp)
    {
    const char *pcmd = pftp->buf + pftp->rptr;
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
#endif
    log_printf ("ftpc_cmnd: $");
    log_line (pcmd);
    if ( ! strncasecmp (pcmd, "user", 4) )
        {
        pftp->psta = sta230;
        }
    else if ( ! strncasecmp (pcmd, "quit", 4) )
        {
        pftp->psta = sta221;
        ftpc_quit (pftp);
        return;
        }
    else if ( ! strncasecmp (pcmd, "port", 4) )
        {
        int n = 0;
        word b = 0;
        BOOLEAN bValid = TRUE;
        pcmd += 4;
        while ( *pcmd == ' ' ) ++pcmd;
        while (TRUE)
            {
            char ch = *pcmd;
            if ((ch >= '0') && (ch <= '9'))
                {
                b = 10 * b + ( ch - '0' );
                if ( b > 255 ) bValid = FALSE;
                }
            else if (ch == ',')
                {
                pftp->daddr[n] = (byte) b;
                b = 0;
                if ( ++n >= 6 )
                    {
                    bValid = FALSE;
                    break;
                    }
                }
            else
                {
                break;
                }
            ++pcmd;
            }
        if ( n == 5 ) pftp->daddr[n] = (byte) b;
        else bValid = FALSE;
        if ( bValid )
            {
            if ( pftp->pasv != 0 )
                {
                net_disconnect (pftp->sadat.socket, (NF) ftp_null, pftp);
                pftp->pasv = 0;
                }
            pftp->psta = sta200;
            }
        else
            {
            pftp->psta = sta501;
            }
        }
    else if ( ! strncasecmp (pcmd, "pasv", 4) )
        {
        byte addr[6];
        char *prep;
        int i;
        if ( pftp->pasv == 0 )
            {
            pftp->pasv = net_next_ephemeral_port ();
            net_listen (pftp->sadat.socket, pftp->pasv, (NF) ftpd_connect, pftp);
            }
        wiz_read_n (W5100_ADDR_REG_SIPR0, addr, 4);
        addr[4] = pftp->pasv >> 8;
        addr[5] = pftp->pasv & 0xFF;
#ifdef DEBUG
        log_printf ("Source = $");
        log_uint (addr[0]);
#endif
        pftp->psta = sta227;
        prep = sta227 + 17;
        *(prep++) = '(';
        prep = ftp_uint (prep, addr[0]);
        for (i = 1; i < 6; ++i)
            {
            *(prep++) = ',';
            prep = ftp_uint (prep, addr[i]);
#ifdef DEBUG
            log_printf (".$");
            log_uint (addr[i]);
#endif
            }
#ifdef DEBUG
        log_printf ("\r\n$");
#endif
        *(prep++) = ')';
        *(prep++) = '\r';
        *(prep++) = '\n';
        *(prep++) = '\0';
        }
    else if ( ! strncasecmp (pcmd, "retr", 4) )
        {
        if ( ftp_fname (pcmd + 4, &pftp->fcb) )
            {
            pftp->fcb.cr = 0;
            if ( bdos_open_file (&pftp->fcb) < 4 )
                {
                pftp->md = mdSend;
                ftpd_start (pftp);
                }
            else
                {
                pftp->psta = sta550;
                }
            }
        else pftp->psta = sta501;
        }
    else if ( ! strncasecmp (pcmd, "stor", 4) )
        {
        if (ftp_fname (pcmd + 4, &pftp->fcb) )
            {
            pftp->fcb.cr = 0;
            bdos_delete_files (&pftp->fcb);
            if ( bdos_create_file (&pftp->fcb) < 4 )
                {
                pftp->md = mdRecv;
                pftp->ndat = 0;
                ftpd_start (pftp);
                }
            else
                {
                pftp->psta = sta550;
                }
            }
        else pftp->psta = sta501;
        }
    else if ( ! strncasecmp (pcmd, "dele", 4) )
        {
        if (ftp_fname (pcmd + 4, &pftp->fcb) )
            {
            if ( bdos_delete_files (&pftp->fcb) < 4 ) pftp->psta = sta250;
            else pftp->psta = sta550;
            }
        else pftp->psta = sta501;
        }
    else if ( ! strncasecmp (pcmd, "rnfr", 4) )
        {
        if (ftp_fname (pcmd + 4, &pftp->fcb) ) pftp->psta = sta350;
        else pftp->psta = sta501;
        }
    else if ( ! strncasecmp (pcmd, "rnto", 4) )
        {
        if ( ftp_fname (pcmd + 4, (FCB *) pftp->fcb.alb) )
            {
            if ( bdos_rename_file (&pftp->fcb) ) pftp->psta = sta250;
            else pftp->psta = sta550;
            }
        else pftp->psta = sta501;
        }
    else if (( ! strncasecmp (pcmd, "nlst", 4) ) || ( ! strncasecmp (pcmd, "list", 4) )
        || ( ! strncasecmp (pcmd, "mlsd", 4) ))
        {
        BOOLEAN bValid;
        if ( pcmd[4] == ' ' )
            {
            bValid = ftp_fname (pcmd + 4, &pftp->fcb);
            }
        else
            {
            pftp->fcb.drive = 0;
            strcpy (pftp->fcb.filename, "????????");
            strcpy (pftp->fcb.type, "???");
            bValid = TRUE;
            }
        if ( bValid )
            {
            ftp_dlist ();
            pftp->md = mdList;
            pftp->iFind = 0;
            if (( *pcmd == 'L' ) || ( *pcmd == 'l' )) pftp->md |= mdLDir;
            else if (( *pcmd == 'M' ) || ( *pcmd == 'm' )) pftp->md |= mdMLst;
            ftpd_start (pftp);
            }
        else pftp->psta = sta501;
        }
    else if ( ! strncasecmp (pcmd, "type", 4) )
        {
        const char *ps = pcmd + 4;
        pftp->psta = sta200;
        while (*ps == ' ') ++ps;
        if ( strchr ("AaIi", *ps) == NULL )
            {
            pftp->psta = sta504;
            }
        else
            {
            ++ps;
            while (*ps == ' ') ++ps;
            if ( strchr ("NnTt\r", *ps) == NULL ) pftp->psta = sta504;
            }
        }
    else if ( ! strncasecmp (pcmd, "mode", 4) )
        {
        const char *ps = pcmd + 4;
        while (*ps == ' ') ++ps;
        if ( strchr ("Ss", *ps) != NULL ) pftp->psta = sta200;
        else pftp->psta = sta504;
        }
    else if ( ! strncasecmp (pcmd, "stru", 4) )
        {
        const char *ps = pcmd + 4;
        while (*ps == ' ') ++ps;
        if ( strchr ("Ff", *ps) != NULL ) pftp->psta = sta200;
        else pftp->psta = sta504;
        }
    else if (( ! strncasecmp (pcmd, "pwd", 3) ) || ( ! strncasecmp (pcmd, "xpwd", 4) ))
        {
        sta257[5] = 'A' + bdos_current_drive ();
        pftp->psta = sta257;
        }
    else if ( ! strncasecmp (pcmd, "cwd", 3) )
        {
        const char *ps = pcmd + 4;
        while (*ps == ' ') ++ps;
        pftp->psta = sta501;
        if ( ( ps[1] == ':' )
            && ( ( ps[2] == '/' ) || ( ps[2] == '\\' ) )
            && ( ps[3] == '\r' ) )
            {
            char ch = *ps;
            if ( ( ch >= 'a' ) && ( ch <= 'z' ) ) ch -= 0x20;
            ch -= 'A';
            if ( ch < 16 )
                {
                word drvs = bdos_logged_drives ();
                if ( drvs & ( 1 << ch ) )
                    {
                    bdos_select_drive (ch);
                    pftp->psta = sta250;
                    }
                }
            }
        else if ( ( ( ps[0] == '/' ) || ( ps[0] == '\\' ) )
            && ( ps[1] == '\r' ) )
            {
            pftp->psta = sta250;
            }
        }
    else if ( ! strncasecmp (pcmd, "feat", 4) )
        {
        pftp->psta = sta211;
        }
    else if ( ! strncasecmp (pcmd, "opts", 4) )
        {
        pftp->psta = sta501;
        }
    else if ( ! strncasecmp (pcmd, "syst", 4) )
        {
        pftp->psta = sta215;
        }
    else
        {
        pftp->psta = sta500;
        }
    ftpc_reply (pftp);
    }

static void ftpc_read(FTP *pftp)
    {
    const char *pend;
    pftp->buf[pftp->wptr] = '\0';
    pend = strstr (&pftp->buf[pftp->rptr], "\r\n");
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
	log_printf("ftpc_read: rptr = $");
    log_uint (pftp->rptr);
    log_printf(" wptr = $");
    log_uint (pftp->wptr);
    log_printf(" pend = $");
    log_word ((word) pend);
    log_printf("\r\n");
#endif
    if ( pend != NULL )
        {
        ftpc_cmnd(pftp);
        pftp->rptr = pend + 2 - pftp->buf;
        }
    else
        {
        if ( pftp->wptr > pftp->rptr ) memcpy (pftp->buf, &pftp->buf[pftp->rptr], pftp->wptr - pftp->rptr);
        pftp->wptr -= pftp->rptr;
        pftp->rptr = 0;
        ftpc_recv (pftp);
        }
    }

static void ftpc_recvd(FTP *pftp)
    {
    SOCKET *socket = pftp->sacmd.socket;
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
	log_printf("ftpc_recvd: $");
    ftp_status (pftp->sacmd.socket);
#endif
    if ( socket->state == SS_ESTABLISHED )
        {
        pftp->wptr += socket->actual;
        pftp->buf[pftp->wptr] = '\0';
        ftpc_read (pftp);
        }
    else
        {
        ftpc_close (pftp);
        }
    }

static void ftpc_recv(FTP *pftp)
	{
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
	log_printf("ftpc_recv: $");
    ftp_status (pftp->sacmd.socket);
#endif
	net_recv(pftp->sacmd.socket, pftp->buf + pftp->wptr, FTPC_BLEN - pftp->wptr, (NF)ftpc_recvd, pftp);
	}

static void ftpc_open(FTP *pftp)
    {
    word sra = pftp->sacmd.socket->sra;
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
#endif
    log_printf ("ftpc_open:\r\n$");
    wiz_read_n (sra + W5100_OFFSET_SOCKET_DIPR0, pftp->daddr, 4);
    pftp->daddr[4] = 0;
    pftp->daddr[5] = PORT_FTPD;
    pftp->psta = sta220;
    ftpc_reply(pftp);
    }

static void ftpc_listen(FTP *pftp)
	{
#ifdef DEBUG
    log_head (pftp->sacmd.socket);
	log_printf("ftpc_listen:\r\n$");
#endif
	net_listen(pftp->sacmd.socket, PORT_FTPC, (NF)ftpc_open, pftp);
	}

static void ftpd_alloc (FTP *pftp)
    {
#ifdef DEBUG
	log_printf("ftpd_alloc:\r\n$");
#endif
	net_alloc(&pftp->sadat, (NF) ftpc_listen, pftp);
    }

static void ftp_start(FTP *pftp)
	{
	log_printf("ftpc_start:\r\n$");
    pftp->rptr = 0;
    pftp->wptr = 0;
    pftp->pasv = 0;
    pftp->dcon = 0;
    pftp->md = mdIdle;
    pftp->sacmd.fiber.queued = 0;
    pftp->sadat.fiber.queued = 0;
    pftp->fbr.queued = 0;
	net_alloc(&pftp->sacmd, (NF) ftpd_alloc, pftp);
	}

static void first(void *nfp)
	{
    nfp = nfp;
    ftp_start (&ftp1);
    ftp_start (&ftp2);
    }

static FIBER first_fiber = { NULL, first, NULL };

void quit(void)
    {
    log_printf ("Exiting due to keyboard request\r\n$");
    log_close();
    bdos_exit();
    }

void parse(void)
    {
    const char *ctail = (char *) 0x80;
    const char *cend = ctail + ctail[0] + 1;
    const char *ps = ctail + 1;
    while ( ps < cend )
        {
        if ( *ps == ' ' )
            {
            ++ps;
            }
        else if ( *ps == '-' )
            {
            ++ps;
            switch (*ps)
                {
                case 'L':
                case 'l':
                    ++ps;
                    ps = log_open (ps, cend);
                    break;
                default:
                    log_fatal ("Invalid parameter\r\n$");
                }
            }
        else
            {
            ps = log_open (ps, cend);
            }
        }
    }

int main(void)
	{
    int nshow = 0;
    parse();
	log_printf("File Transfer Protocol Daemon " VERSION "\r\n$");
	fiber_init();
	interrupts_enable();
	net_init();
	fiber_enqueue(&first_fiber);
	for ( ;; )
		{
        ++nLoop;
#ifdef DEBUG
        if ( log_nshow != nshow )
            {
            log_printf ("Main loop\r\n");
            nshow = log_nshow;
            }
#endif
        if ( bdos_inkey() == 'q' ) quit();
		net_handle_interrupts();
		fiber_drain();
		}
	}

