/*

wiz.c - Access to WIZnet on NFX

*/

/*...semail from Martin:0:*/
/*

The NFX uses a Wiznet 5100 chip on an adapter board, it's got the 8 bit IDE
on-board and a custom 40 column CP/M.

The Wiznet is mapped a bit like the VDP it has 4 I/O ports A0 to A3. Control,
address msb, address, lsb and data. Although unlike the VDP you can turn the
auto increment on and off as required. Datasheet and rom source code such as
it is enclosed. The interrupt from the net card isn't connected as I couldn't
see a safe way to do it.
 
The IDE is mapped B0-BF
 
There's 3 general routines in the rom:
FFD9 for net init
FFDC for net write
FFDF for net read
  
The wiznet chip memory maps its registers so within the 32k map, the read and
write routines ought to be able to access both the registers and the data ram.
I did manage to set up some test code using an SDX rom to give the board an
IP address so it could respond to pings from a PC. There's only room for one
paged rom as it's using grom to page to rom. I'd decided because it had more
useable memory, CPM would be a better environment to do any more work so
replaced the SDX rom with a CPM one.

*/
/*...e*/

/*...sincludes:0:*/
#include <stdio.h>
#include "w5100.h"
#include "wizio.h"
#include "wiz.h"
#include "log.h"

/*...vw5100\46\h:0:*/
/*...vwizio\46\h:0:*/
/*...vwiz\46\h:0:*/
/*...e*/

#if DYNAMIC_PORT

/*
SDCC can generate inline IN and OUT instructions for fixed port addresses.
However there is no apparent way of doing so for dynamically assigned addresses.
One option would have been to write subroutines to perform the port read and write
but that would introduce subroutine call overhead and slow down the code.

Instead, the ASM code generated by compiling the fixed port code has been
copied inline and hand editted to replace IN A,(#) and OUT (#), A by IN A,(C) and
OUT (C),A.

The ASM code has also been optimised in other ways, including inlining wiz_set_addr()
and the contents of "wizio.s".
*/
static byte wiz_port_control;
static byte wiz_port_addr_msb;
static byte wiz_port_addr_lsb;
static byte wiz_port_data;

#else
/*...sports:0:*/
#define	NFX_WIZ_CONTROL_PORT  0xa0
#define	NFX_WIZ_ADDR_MSB_PORT 0xa1
#define	NFX_WIZ_ADDR_LSB_PORT 0xa2
#define	NFX_WIZ_DATA_PORT     0xa3
	
__sfr __at NFX_WIZ_CONTROL_PORT  NFX_WIZ_CONTROL ;
__sfr __at NFX_WIZ_ADDR_MSB_PORT NFX_WIZ_ADDR_MSB;
__sfr __at NFX_WIZ_ADDR_LSB_PORT NFX_WIZ_ADDR_LSB;
__sfr __at NFX_WIZ_DATA_PORT     NFX_WIZ_DATA    ;
/*...e*/

#define	USE_WIZIO
#endif

#if DYNAMIC_PORT
void wiz_config (byte port)
    {
    byte pt = port;		// Generates slightly more efficient code than just using port
    wiz_port_control = pt;
    wiz_port_addr_msb = ++pt;
    wiz_port_addr_lsb = ++pt;
    wiz_port_data = ++pt;
    }
#endif

/*...swiz_reset:0:*/
void wiz_reset(void)
    {
#if DYNAMIC_PORT
    __asm
;wiz.c:80: NFX_WIZ_CONTROL = W5100_MR_RESET | W5100_MR_INDIRECT | W5100_MR_INCREMENT;
	ld	a, #0x83
;	out	(_NFX_WIZ_CONTROL), a
	ld	c, (_wiz_port_control)
	out	(c), a
;wiz.c:82: while ( NFX_WIZ_CONTROL & W5100_MR_RESET )
00101$:
;	in	a, (_NFX_WIZ_CONTROL)
	in	a, (c)
	rlca
	jr	C,00101$
;wiz.c:91: NFX_WIZ_CONTROL = W5100_MR_INDIRECT | W5100_MR_INCREMENT;
	ld	a, #0x03
;	out	(_NFX_WIZ_CONTROL), a
	out	(c), a
;wiz.c:92: }
	ret
	__endasm;
#else
    NFX_WIZ_CONTROL = W5100_MR_RESET | W5100_MR_INDIRECT | W5100_MR_INCREMENT;
    /* Wait for reset to complete */
    while ( NFX_WIZ_CONTROL & W5100_MR_RESET )
        {
#ifdef DEBUG
        log_printf ("NFX_WIZ_CONTROL = $");
        log_byte (NFX_WIZ_CONTROL);
        log_printf ("\r\n$");
#endif
        }
    /* note: don't use W5100_MR_RESET here */
    NFX_WIZ_CONTROL = W5100_MR_INDIRECT | W5100_MR_INCREMENT;
#endif
    }
/*...e*/
/*...swiz_set_addr:0:*/
void wiz_set_addr(word addr)
    {
#if DYNAMIC_PORT
    addr;
    __asm
;wiz.c:97: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:98: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
;wiz.c:99: }
	ret
	__endasm;
#else
    NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
    NFX_WIZ_ADDR_LSB = (byte)  addr    ;
#endif
    }
/*...e*/
/*...swiz_write:0:*/
void wiz_write(word addr, byte value)
    {
#if DYNAMIC_PORT
    addr;
    value;
    __asm
;wiz.c:169: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:170: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
;wiz.c:171: NFX_WIZ_DATA = value;
	ld	a, 2 (iy)
;	out	(_NFX_WIZ_DATA), a
	inc	c
	inc	c
	out	(c), a
;wiz.c:173: }
	ret
	__endasm;
#else
//    wiz_set_addr(addr);
    NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
    NFX_WIZ_ADDR_LSB = (byte)  addr    ;
    NFX_WIZ_DATA = value;
#endif
    }
/*...e*/
/*...swiz_write_word:0:*/
void wiz_write_word(word addr, word ptr)
    {
#if DYNAMIC_PORT
    addr;
    ptr;
    __asm
;wiz.c:183: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:184: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
;wiz.c:185: NFX_WIZ_DATA = (byte) (ptr>>8);
	ld	a, 3 (iy)
;	out	(_NFX_WIZ_DATA), a
	inc	c
	inc	c
	out	(c), a
;wiz.c:186: NFX_WIZ_DATA = (byte)  ptr    ;
	ld	a, 2 (iy)
;	out	(_NFX_WIZ_DATA), a
	out	(c), a
;wiz.c:188: }
	ret
	__endasm;
#else
//    wiz_set_addr(addr);
    NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
    NFX_WIZ_ADDR_LSB = (byte)  addr    ;
    NFX_WIZ_DATA = (byte) (ptr>>8);
    NFX_WIZ_DATA = (byte)  ptr    ;
#endif
    }
/*...e*/
/*...swiz_write_n:0:*/
void wiz_write_n(word addr, const byte *values, word n)
    {
#if DYNAMIC_PORT
    addr;
    values;
    n;
    __asm
;wiz.c:183: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:184: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
	inc	c
	inc	c
;wiz.c:195: wizio_out(values, n, NFX_WIZ_DATA_PORT);
	ld	l,2(iy)
	ld	h,3(iy)
	ld	e,4(iy)
	ld	d,5(iy)
	inc	d
	dec	d
	jr	z,00002$
	ld	b,#0
00001$:
	otir
	dec	d
	jr	nz,00001$
00002$:
	inc	e
	dec	e
	ret	z
	ld	b,e
	otir
	ret
	__endasm;
#else
    wiz_set_addr(addr);
#ifdef USE_WIZIO
    wizio_out(values, n, NFX_WIZ_DATA_PORT);
#else
    while ( n-- > 0 )
	NFX_WIZ_DATA = *values++;
#endif
#endif
    }
/*...e*/
/*...swiz_read:0:*/
byte wiz_read(word addr)
    {
#if DYNAMIC_PORT
    addr;
    __asm
;wiz.c:210: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:211: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
;wiz.c:212: return NFX_WIZ_DATA;
;	in	a, (_NFX_WIZ_DATA)
;	ld	l, a
	inc	c
	inc	c
	in	l, (c)
;wiz.c:214: }
	ret
	__endasm;
#else
//    wiz_set_addr(addr);
    NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
    NFX_WIZ_ADDR_LSB = (byte)  addr    ;
    return NFX_WIZ_DATA;
#endif
    }
/*...e*/
/*...swiz_read_word:0:*/
word wiz_read_word(word addr)
    {
#if DYNAMIC_PORT
    addr;
    __asm
;wiz.c:210: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:211: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
;wiz.c:227: r0 = NFX_WIZ_DATA;
;wiz.c:228: r1 = NFX_WIZ_DATA;
;wiz.c:229: return ((word)r0)<<8 | (word)r1;
	inc	c
	inc	c
	in	l, (c)
	in	h, (c)
;wiz.c:231: }
	ret
	__endasm;
#else
    byte r0, r1;
//    wiz_set_addr(addr);
    NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
    NFX_WIZ_ADDR_LSB = (byte)  addr    ;
    r0 = NFX_WIZ_DATA;
    r1 = NFX_WIZ_DATA;
    return ((word)r0)<<8 | (word)r1;
#endif
    }
/*...e*/
/*...swiz_read_n:0:*/
void wiz_read_n(word addr, byte *values, word n)
    {
#if DYNAMIC_PORT
    addr;
    values;
    n;
    __asm
;wiz.c:210: NFX_WIZ_ADDR_MSB = (byte) (addr>>8);
	ld	iy, #2
	add	iy, sp
	ld	a, 1 (iy)
;	out	(_NFX_WIZ_ADDR_MSB), a
	ld	c, (_wiz_port_addr_msb)
	out	(c), a
;wiz.c:211: NFX_WIZ_ADDR_LSB = (byte)  addr    ;
	ld	a, 0 (iy)
;	out	(_NFX_WIZ_ADDR_LSB), a
	dec	c
	out	(c), a
	inc	c
	inc	c
;wiz.c:238: wizio_in(values, n, NFX_WIZ_DATA_PORT);
	ld	l,2(iy)
	ld	h,3(iy)
	ld	e,4(iy)
	ld	d,5(iy)
	inc	d
	dec	d
	jr	z,00002$
	ld	b,#0
00001$:
	inir
	dec	d
	jr	nz,00001$
00002$:
	inc	e
	dec	e
	ret	z
	ld	b,e
	inir
	ret
	__endasm;
#else
    wiz_set_addr(addr);
#ifdef USE_WIZIO
    wizio_in(values, n, NFX_WIZ_DATA_PORT);
#else
    while ( n-- > 0 )
	*values++ = NFX_WIZ_DATA;
#endif
#endif
    }
/*...e*/
